@@ -27,6 +27,8 @@
 #include "defs.h"
 
 #define USE_SAFE_CHDIR 1
+#undef  STAT_MOUNTPOINTS
+
 
 #include <errno.h>
 #include <assert.h>
@@ -73,8 +75,10 @@
 #define apply_predicate(pathname, stat_buf_ptr, node)	\
   (*(node)->pred_func)((pathname), (stat_buf_ptr), (node))
 
-
+#ifdef STAT_MOUNTPOINTS
 static void init_mounted_dev_list(void);
+#endif
+
 static void process_top_path PARAMS((char *pathname, mode_t mode));
 static int process_path PARAMS((char *pathname, char *name, boolean leaf, char *parent, mode_t type));
 static void process_dir PARAMS((char *pathname, char *name, int pathlen, struct stat *statp, char *parent));
@@ -127,6 +131,13 @@ enum TraversalDirection
     TraversingDown
   };
 
+enum WdSanityCheckFatality
+  {
+    FATAL_IF_SANITY_CHECK_FAILS,
+    RETRY_IF_SANITY_CHECK_FAILS,
+    NON_FATAL_IF_SANITY_CHECK_FAILS
+  };
+
 
 int
 following_links(void)
@@ -380,6 +391,19 @@ main (int argc, char **argv)
   int end_of_leading_options = 0; /* First arg after any -H/-L etc. */
   program_name = argv[0];
 
+  /* We call check_nofollow() before setlocale() because the numbers 
+   * for which we check (in the results of uname) definitiely have "."
+   * as the decimal point indicator even under locales for which that 
+   * is not normally true.   Hence atof() would do the wrong thing 
+   * if we call it after setlocale().
+   */
+#ifdef O_NOFOLLOW
+  options.open_nofollow_available = check_nofollow();
+#else
+  options.open_nofollow_available = false;
+#endif
+
+  
 #ifdef HAVE_SETLOCALE
   setlocale (LC_ALL, "");
 #endif
@@ -419,12 +443,6 @@ main (int argc, char **argv)
   else
     options.output_block_size = 1024;
 
-#ifdef O_NOFOLLOW
-  options.open_nofollow_available = check_nofollow();
-#else
-  options.open_nofollow_available = false;
-#endif
-  
   if (getenv("FIND_BLOCK_SIZE"))
     {
       error (1, 0, _("The environment variable FIND_BLOCK_SIZE is not supported, the only thing that affects the block size is the POSIXLY_CORRECT environment variable"));
@@ -433,21 +451,6 @@ main (int argc, char **argv)
   options.no_leaf_check = false;
   set_follow_state(SYMLINK_NEVER_DEREF); /* The default is equivalent to -P. */
 
-  /* safely_chdir() needs to check that it has ended up in the right place. 
-   * To avoid bailing out when something gets automounted, it checks if 
-   * the target directory appears to have had a directory mounted on it as
-   * we chdir()ed.  The problem with this is that in order to notice that 
-   * a filesystem was mounted, we would need to lstat() all the mount points.
-   * That strategy loses if our machine is a client of a dead NFS server.
-   *
-   * Hence if safely_chdir() and wd_sanity_check() can manage without needing 
-   * to know the mounted device list, we do that.  
-   */
-  if (!options.open_nofollow_available)
-    {
-      init_mounted_dev_list();
-    }
-  
 #ifdef DEBUG
   fprintf (stderr, "cur_day_start = %s", ctime (&options.cur_day_start));
 #endif /* DEBUG */
@@ -586,6 +589,24 @@ main (int argc, char **argv)
   fprintf(stderr, "\n");
 #endif /* DEBUG */
 
+  /* safely_chdir() needs to check that it has ended up in the right place. 
+   * To avoid bailing out when something gets automounted, it checks if 
+   * the target directory appears to have had a directory mounted on it as
+   * we chdir()ed.  The problem with this is that in order to notice that 
+   * a filesystem was mounted, we would need to lstat() all the mount points.
+   * That strategy loses if our machine is a client of a dead NFS server.
+   *
+   * Hence if safely_chdir() and wd_sanity_check() can manage without needing 
+   * to know the mounted device list, we do that.  
+   */
+  if (!options.open_nofollow_available)
+    {
+#ifdef STAT_MOUNTPOINTS
+      init_mounted_dev_list();
+#endif
+    }
+  
+
   starting_desc = open (".", O_RDONLY);
   if (0 <= starting_desc && fchdir (starting_desc) != 0)
     {
@@ -656,6 +677,20 @@ specific_dirname(const char *dir)
     }
 }
 
+
+
+/* Return non-zero if FS is the name of a filesystem that is likely to
+ * be automounted
+ */
+static int
+fs_likely_to_be_automounted(const char *fs)
+{
+  return ( (0==strcmp(fs, "nfs")) || (0==strcmp(fs, "autofs")) || (0==strcmp(fs, "subfs")));
+}
+
+
+
+#ifdef STAT_MOUNTPOINTS
 static dev_t *mounted_devices = NULL;
 static size_t num_mounted_devices = 0u;
 
@@ -722,21 +757,63 @@ get_mount_state(dev_t newdev)
     return MountPointRecentlyUnmounted;
 }
 
-
-/* Return non-zero if FS is the name of a filesystem that is likely to
- * be automounted
+
+
+/* We stat()ed a directory, chdir()ed into it (we know this 
+ * since direction is TraversingDown), stat()ed it again,
+ * and noticed that the device numbers are different.  Check
+ * if the filesystem was recently mounted. 
+ * 
+ * If it was, it looks like chdir()ing into the directory
+ * caused a filesystem to be mounted.  Maybe automount is
+ * running.  Anyway, that's probably OK - but it happens
+ * only when we are moving downward.
+ *
+ * We also allow for the possibility that a similar thing
+ * has happened with the unmounting of a filesystem.  This
+ * is much rarer, as it relies on an automounter timeout
+ * occurring at exactly the wrong moment.
  */
-static int
-fs_likely_to_be_automounted(const char *fs)
+static enum WdSanityCheckFatality
+dirchange_is_fatal(const char *specific_what,
+		   enum WdSanityCheckFatality isfatal,
+		   int silent,
+		   struct stat *newinfo)
 {
-  return ( (0==strcmp(fs, "nfs")) || (0==strcmp(fs, "autofs")) || (0==strcmp(fs, "subfs")));
+  enum MountPointStateChange transition = get_mount_state(newinfo->st_dev);
+  switch (transition)
+    {
+    case MountPointRecentlyUnmounted:
+      isfatal = NON_FATAL_IF_SANITY_CHECK_FAILS;
+      if (!silent)
+	{
+	  error (0, 0,
+		 _("Warning: filesystem %s has recently been unmounted."),
+		 specific_what);
+	}
+      break;
+	      
+    case MountPointRecentlyMounted:
+      isfatal = NON_FATAL_IF_SANITY_CHECK_FAILS;
+      if (!silent)
+	{
+	  error (0, 0,
+		 _("Warning: filesystem %s has recently been mounted."),
+		 specific_what);
+	}
+      break;
+
+    case MountPointStateUnchanged:
+      /* leave isfatal as it is */
+      break;
+    }
+  
+  return isfatal;
 }
 
-enum WdSanityCheckFatality
-  {
-    FATAL_IF_SANITY_CHECK_FAILS,
-    NON_FATAL_IF_SANITY_CHECK_FAILS
-  };
+
+#endif
+
 
 
 /* Examine the results of the stat() of a directory from before we
@@ -795,82 +872,52 @@ wd_sanity_check(const char *thing_to_stat,
       specific_what = specific_dirname(what);
       fstype = filesystem_type(newinfo);
       silent = fs_likely_to_be_automounted(fstype);
-      
+
       /* This condition is rare, so once we are here it is 
        * reasonable to perform an expensive computation to 
        * determine if we should continue or fail. 
        */
       if (TraversingDown == direction)
 	{
-	  /* We stat()ed a directory, chdir()ed into it (we know this 
-	   * since direction is TraversingDown), stat()ed it again,
-	   * and noticed that the device numbers are different.  Check
-	   * if the filesystem was recently mounted. 
-	   * 
-	   * If it was, it looks like chdir()ing into the directory
-	   * caused a filesystem to be mounted.  Maybe automount is
-	   * running.  Anyway, that's probably OK - but it happens
-	   * only when we are moving downward.
-	   *
-	   * We also allow for the possibility that a similar thing
-	   * has happened with the unmounting of a filesystem.  This
-	   * is much rarer, as it relies on an automounter timeout
-	   * occurring at exactly the wrong moment.
-	   */
-	  enum MountPointStateChange transition = get_mount_state(newinfo->st_dev);
-	  switch (transition)
-	    {
-	    case MountPointRecentlyUnmounted:
-	      isfatal = NON_FATAL_IF_SANITY_CHECK_FAILS;
-	      if (!silent)
-		{
-		  error (0, 0,
-			 _("Warning: filesystem %s has recently been unmounted."),
-			 specific_what);
-		}
-	      break;
-	      
-	    case MountPointRecentlyMounted:
-	      isfatal = NON_FATAL_IF_SANITY_CHECK_FAILS;
-	      if (!silent)
-		{
-		  error (0, 0,
-			 _("Warning: filesystem %s has recently been mounted."),
-			 specific_what);
-		}
-	      break;
-
-	    case MountPointStateUnchanged:
-	      /* leave isfatal as it is */
-	      break;
-	    }
+#ifdef STAT_MOUNTPOINTS
+	  isfatal = dirchange_is_fatal(specific_what,isfatal,silent,newinfo);
+#else
+	  isfatal = RETRY_IF_SANITY_CHECK_FAILS;
+#endif
 	}
 
-      if (FATAL_IF_SANITY_CHECK_FAILS == isfatal)
+      switch (isfatal)
 	{
-	  fstype = filesystem_type(newinfo);
-	  error (1, 0,
-		 _("%s%s changed during execution of %s (old device number %ld, new device number %ld, filesystem type is %s) [ref %ld]"),
-		 specific_what,
-		 parent ? "/.." : "",
-		 program_name,
-		 (long) old_dev,
-		 (long) newinfo->st_dev,
-		 fstype,
-		 line_no);
-	  /*NOTREACHED*/
+	case FATAL_IF_SANITY_CHECK_FAILS:
+	  {
+	    fstype = filesystem_type(newinfo);
+	    error (1, 0,
+		   _("%s%s changed during execution of %s (old device number %ld, new device number %ld, filesystem type is %s) [ref %ld]"),
+		   specific_what,
+		   parent ? "/.." : "",
+		   program_name,
+		   (long) old_dev,
+		   (long) newinfo->st_dev,
+		   fstype,
+		   line_no);
+	    /*NOTREACHED*/
+	    return false;
+	  }
+	  
+	case NON_FATAL_IF_SANITY_CHECK_FAILS:
+	  {
+	    /* Since the device has changed under us, the inode number 
+	     * will almost certainly also be different. However, we have 
+	     * already decided that this is not a problem.  Hence we return
+	     * without checking the inode number.
+	     */
+	    free(specific_what);
+	    return true;
+	  }
+
+	case RETRY_IF_SANITY_CHECK_FAILS:
 	  return false;
 	}
-      else
-	{
-	  /* Since the device has changed under us, the inode number 
-	   * will almost certainly also be different. However, we have 
-	   * already decided that this is not a problem.  Hence we return
-	   * without checking the inode number.
-	   */
-	  free(specific_what);
-	  return true;
-	}
     }
 
   /* Device number was the same, check if the inode has changed. */
@@ -925,10 +972,19 @@ safely_chdir_lstat(const char *dest,
   int rv, dotfd=-1;
   int saved_errno;		/* specific_dirname() changes errno. */
   boolean rv_set = false;
+  int tries = 0;
+  enum WdSanityCheckFatality isfatal = RETRY_IF_SANITY_CHECK_FAILS;
   
   saved_errno = errno = 0;
 
   dotfd = open(".", O_RDONLY);
+
+  /* We jump back to here if wd_sanity_check()
+   * recoverably triggers an alert.
+   */
+ retry:
+  ++tries;
+  
   if (dotfd >= 0)
     {
       /* Stat the directory we're going to. */
@@ -998,13 +1054,45 @@ safely_chdir_lstat(const char *dest,
 	    {
 	      /* check we ended up where we wanted to go */
 	      boolean changed = false;
-	      wd_sanity_check(".", program_name, ".",
-			      statbuf_dest->st_dev,
-			      statbuf_dest->st_ino,
-			      &statbuf_arrived, 
-			      0, __LINE__, direction,
-			      FATAL_IF_SANITY_CHECK_FAILS,
-			      &changed);
+	      if (!wd_sanity_check(".", program_name, ".",
+				   statbuf_dest->st_dev,
+				   statbuf_dest->st_ino,
+				   &statbuf_arrived, 
+				   0, __LINE__, direction,
+				   isfatal,
+				   &changed))
+		{
+		  /* Only allow one failure. */
+		  if (RETRY_IF_SANITY_CHECK_FAILS == isfatal)
+		    {
+		      if (0 == fchdir(dotfd))
+			{
+			  isfatal = FATAL_IF_SANITY_CHECK_FAILS;
+			  goto retry;
+			}
+		      else
+			{
+			  /* Failed to return to original directory,
+			   * but we know that the current working
+			   * directory is not the one that we intend
+			   * to be in.  Since fchdir() failed, we
+			   * can't recover from this and so this error
+			   * is fatal.
+			   */
+			  error(1, errno,
+				"failed to return to parent directory");
+			}
+		    }
+		  else
+		    {
+		      /* XXX: not sure what to use as an excuse here. */
+		      rv = SafeChdirFailNonexistent;
+		      rv_set = true;
+		      saved_errno = 0;
+		      goto fail;
+		    }
+		}
+	      
 	      close(dotfd);
 	      return SafeChdirOK;
 	    }
@@ -1080,10 +1168,11 @@ safely_chdir_lstat(const char *dest,
 }
 
 #if defined(O_NOFOLLOW)
-/* Safely change working directory to the specified subdirectory.
- * We use open() with O_NOFOLLOW, followed by fchdir().  This ensures
- * that we don't follow symbolic links (of course, we do follow them
- * if the -L option is in effect).
+/* Safely change working directory to the specified subdirectory.  If
+ * we are not allowed to follow symbolic links, we use open() with
+ * O_NOFOLLOW, followed by fchdir().  This ensures that we don't
+ * follow symbolic links (of course, we do follow them if the -L
+ * option is in effect).
  */
 static enum SafeChdirStatus
 safely_chdir_nofollow(const char *dest,
@@ -1091,10 +1180,25 @@ safely_chdir_nofollow(const char *dest,
 		      struct stat *statbuf_dest,
 		      enum ChdirSymlinkHandling symlink_handling)
 {
-  int extraflags = following_links() ? O_NOFOLLOW : 0;
-
+  int extraflags, fd;
+  extraflags = 0;
+  
+  switch (symlink_handling)
+    {
+    case SymlinkFollowOk:
+      extraflags = 0;
+      break;
+      
+    case SymlinkHandleDefault:
+      if (following_links())
+	extraflags = 0;
+      else
+	extraflags = O_NOFOLLOW;
+      break;
+    }
+  
   errno = 0;
-  int fd = open(dest, O_RDONLY|extraflags);
+  fd = open(dest, O_RDONLY|extraflags);
   if (fd < 0)
     {
       switch (errno)
@@ -1171,11 +1275,13 @@ chdir_back (void)
       fprintf(stderr, "chdir_back(): chdir(\"%s\")\n", starting_dir);
 #endif
       
+#ifdef STAT_MOUNTPOINTS
       /* We will need the mounted device list.  Get it now if we don't
        * already have it.
        */
       if (NULL == mounted_devices)
 	init_mounted_dev_list();
+#endif
       
       if (chdir (starting_dir) != 0)
 	error (1, errno, "%s", starting_dir);
@@ -1219,7 +1325,8 @@ process_top_path (char *pathname, mode_t mode)
   state.curdepth = 0;
   state.path_length = strlen (pathname);
 
-  if (0 == strcmp(pathname, parent_dir))
+  if (0 == strcmp(pathname, parent_dir)
+      || 0 == strcmp(parent_dir, "."))
     {
       dirchange = 0;
       base = pathname;
@@ -1436,7 +1543,7 @@ process_path (char *pathname, char *name, boolean leaf, char *parent,
   /* From here on, we're working on a directory.  */
 
   
-  /* Now we really need to stat the directory, even if we knoe the
+  /* Now we really need to stat the directory, even if we know the
    * type, because we need information like struct stat.st_rdev.
    */
   if (get_statinfo(pathname, name, &stat_buf) != 0)
@@ -1747,7 +1854,6 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 	{
 	  enum SafeChdirStatus status;
 	  struct dir_id did;
-	  boolean changed = false;
 	  
 	  /* We could go back and do the next command-line arg
 	     instead, maybe using longjmp.  */
